import json

from deprecated import deprecated
from ase import Atoms
from clease import basis_function as bf
from .concentration import Concentration
from .settings import ClusterExpansionSettings
from .settings_bulk import CEBulk, CECrystal
from .settings_slab import CESlab

__all__ = ("settings_from_json", "settingsFromJSON")


def settings_from_json(fname):
    """Initialize settings from JSON.

    Exists due to compatibility. You should instead use
    `ClusterExpansionSettings.load(fname)`

    Parameters:

    fname: str
        JSON file where settings are stored
    """
    return ClusterExpansionSettings.load(fname)  # pylint: disable=no-member


def old_settings_from_json(fname):
    """
    Initialise settings from JSON file.
    Used for reading old json files from versions < 0.10.2

    Parameters:

    fname: str
        JSON file where settings are stored
    """
    with open(fname, "r") as infile:
        data = json.load(infile)

    factory = data["kwargs"].pop("factory")
    kwargs = data["kwargs"]
    conc = Concentration.from_dict(kwargs["concentration"])
    kwargs["concentration"] = conc
    if factory == "CEBulk":
        settings = CEBulk(**kwargs)
    elif factory == "CECrystal":
        settings = CECrystal(**kwargs)
    elif factory == "CESlab":
        cnv_cell_dict = kwargs.pop("conventional_cell")
        cnv_cell = Atoms.fromdict(cnv_cell_dict)
        kwargs["conventional_cell"] = cnv_cell
        settings = CESlab(**kwargs)
    else:
        raise ValueError(f"Unknown factory {factory}")
    settings.include_background_atoms = data["include_background_atoms"]
    settings.skew_threshold = data["skew_threshold"]
    bf_dict = data["basis_func_type"]
    name = bf_dict.pop("name")

    if name == "polynmial":
        settings.basis_func_type = bf.Polynomial(**bf_dict)
    elif name == "trigonometric":
        settings.basis_func_type = bf.Trigonometric(**bf_dict)
    elif name == "binary_linear":
        settings.basis_func_type = bf.BinaryLinear(**bf_dict)
    return settings


@deprecated(version="0.10.0", reason='You should use "settings_from_json" instead')
def settingsFromJSON(fname):
    return settings_from_json(fname)



from typing import List, Sequence, Dict, Union, Callable, Tuple
import time
import logging

import numpy as np
from numpy.random import choice
from scipy.linalg import solve_triangular

from clease.data_normalizer import DataNormalizer
from clease.tools import split_dataset
from .regression import LinearRegression
from .constrained_ridge import ConstrainedRidge

logger = logging.getLogger(__name__)

__all__ = ("PhysicalRidge",)


class PhysicalRidge(LinearRegression):
    """
    Physical Ridge is a special ridge regression scheme that enforces a
    convergent series. The physical motivation behind the choice of
    prior distributions is motivated by the fact that one expects that
    interactions strengths decays with both the number of atoms in the
    cluster and the diameter of the cluster. See for instance

    Cao, L., Li, C. and Mueller, T., 2018. The use of cluster expansions to
    predict the structures and properties of surfaces and nanostructured
    materials. Journal of chemical information and modeling, 58(12),
    pp.2401-2413.

    This fitting scheme uses Gaussian priors on the coefficients of the model

    P(M) = P_size(M)*P_dia(M), where

    P_size(M) = prod_i exp(-lamb_size*size_decay(size)*coeff_i^2)
    P_dia(M) = prod_i exp(-lamb_dia*dia_decay(dia)*coeff_i^2)

    where size_decay and dia_decay is a monotonically increasing function of
    the size and diameter respectively. The product goes over all coefficients
    in the model M.

    :param lamb_size: Prefactor in front of the size penalization

    :param lamb_dia: Prefactor in fron the the diameter penalization

    :param size_decay: The size_decay function in the priors explained above.
        It can be one of ['linear', 'exponential', 'polyN'],
        where N is any integer, or a callable function with
        the signature f(size), where size is the number of atoms in the
        cluster. If polyN is given the penalization is proportional to
        size**N

    :param dia_decay: The dia_decay function in the priors explained above.
        It can be one of ['linear', 'exponential', 'polyN']
        where N is any integer, of a callable function
        with the signature f(dia) where dia is the diameter.
        If polyN is given the penalization is proportional to dia**N

    :param normalize: If True the data will be normalized to unit variance
        and zero mean before fitting.

        NOTE: Normalization works only when the first column in X corresponds
        to a constant. If the X matrix contains several simultaneous fits
        (e.g. energy, pressure, bulk moduli) there will typically be different
        columns that corresponds to the bias term for the different groups. It
        is recommended to put normalize=False for such cases.
    """

    def __init__(
        self,
        lamb_size: float = 1e-6,
        lamb_dia: float = 1e-6,
        size_decay: Union[str, Callable[[int], float]] = "linear",
        dia_decay: Union[str, Callable[[int], float]] = "linear",
        normalize: bool = True,
    ) -> None:
        super().__init__()
        self.lamb_size = lamb_size
        self.lamb_dia = lamb_dia
        self._size_decay = get_size_decay(size_decay)
        self._dia_decay = get_dia_decay(dia_decay)
        self.sizes = []
        self.diameters = []
        self.normalize = normalize
        self.normalizer = DataNormalizer()
        self._constraint = None

    def add_constraint(self, A: np.ndarray, c: np.ndarray) -> None:
        """
        Adds a constraint that the coefficients (ECI) has to obey,
        A.dot(coeff) = c

        :param A: Matrix describing the linear constraint
        :param c: Vector representing the right hand side of
            constraint equations
        """
        self._constraint = {"A": A, "c": c}

    def fit_data(self, X: np.ndarray, y: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        If normalize is True, a normalized version of the passed data is
        returned. Otherwise, X and y is returned as they are passed.

        :param X: Design matrix
        :param y: Target data
        """
        if self.normalize:
            X_fit, y_fit = self.normalizer.normalize(X, y)
            # Skip constant term since y is shifted to have mean zero
            # (and thus the constant term is zero)
            return X_fit[:, 1:], y_fit
        return X, y

    @property
    def size_decay(self) -> Callable[[int], float]:
        return self._size_decay

    @size_decay.setter
    def size_decay(self, decay: Union[str, Callable[[int], float]]) -> None:
        self._size_decay = get_size_decay(decay)

    @property
    def dia_decay(self) -> Callable[[int], float]:
        return self._dia_decay

    @dia_decay.setter
    def dia_decay(self, decay: Union[str, Callable[[int], float]]) -> None:
        self._dia_decay = get_dia_decay(decay)

    def sizes_from_names(self, names: List[str]) -> None:
        """
        Extract the sizes from a list of correlation function names

        :param names: List of cluster names.
            The length of the list has to match the
            number of columns in the X matrix passed to the fit method.
            Ex: ['c0', 'c1_1', 'c2_d0000_0_00']
        """
        self.sizes = [int(n[1]) for n in names]

    def diameters_from_names(self, names: List[str]) -> None:
        """
        Extract the diameters from a list of correltion function names

        :param names: List of cluster names.
            The length of the list has to match the
            number of columns in the X matrix passed to the fit method.
            Ex: ['c0', 'c1_1', 'c2_d0000_0_00']
        """
        diameters = []
        for n in names:
            if n[1] == "0" or n[1] == "1":
                diameters.append(0.0)
            else:
                dia_str = n.split("_")[1]
                dia = int(dia_str[1:])
                diameters.append(dia)
        self.diameters = diameters

    def fit(self, X: np.ndarray, y: np.ndarray) -> np.ndarray:
        """
        Fit ECIs

        :param X: Design matrix with correlation functions. The shape is N x M,
            where N is the number of data points and M is the number of
            correlation functions

        :param y: Vector with target values. The length of this vector is N
            (e.g. equal to the number of rows in X)
        """
        if X.shape[1] != len(self.sizes) or X.shape[1] != len(self.diameters):
            msg = "Inconsistent number of columns in X. "
            msg += f"Num. columns: {X.shape[1]} "
            msg += f"Num. sizes: {len(self.sizes)} "
            msg += f"Num. diameters: {len(self.diameters)}"
            raise ValueError(msg)

        X_fit, y_fit = self.fit_data(X, y)

        size_decay_func = self.size_decay
        dia_decay_func = self.dia_decay

        size_decay = np.array([size_decay_func(x) for x in self.sizes])
        dia_decay = np.array([dia_decay_func(x) for x in self.diameters])

        penalty = self.lamb_size * size_decay + self.lamb_dia * dia_decay
        if len(penalty) == X_fit.shape[1] + 1:
            # In this case the constant term in X was removed
            penalty = penalty[1:]
        elif len(penalty) != X_fit.shape[1]:
            raise RuntimeError(f"Num. penalty {len(penalty)}. Num feature {X_fit.shape[1]}.")

        if self._constraint is not None:
            # Use the constrained Tikhonov verison
            regressor = ConstrainedRidge(penalty)
            regressor.add_constraint(self._constraint["A"], self._constraint["c"])
            coeff = regressor.fit(X_fit, y_fit)
        else:
            matrix = np.vstack((X_fit, np.diag(np.sqrt(penalty))))
            Q, R = np.linalg.qr(matrix)
            rhs = np.concatenate((y_fit, np.zeros(len(penalty))))
            coeff = solve_triangular(R, Q.T.dot(rhs), lower=False)

        if self.normalize:
            coeff_with_bias_term = np.concatenate(([0.0], coeff))
            coeff_transformed = self.normalizer.convert(coeff_with_bias_term)
            coeff_transformed[0] = self.normalizer.bias(coeff_with_bias_term)
            coeff = coeff_transformed
        return coeff


def linear_size(size: int) -> float:
    if size == 0:
        return 0.0
    return size


def linear_dia(dia: int) -> float:
    return dia


def exponential_size(size: int) -> float:
    if size == 0:
        return 0.0
    return np.exp(size - 1) - 1


def exponential_dia(dia: int) -> float:
    return np.exp(dia) - 1.0


def get_size_decay(decay: Union[str, Callable[[int], float]]) -> Callable[[int], float]:
    if isinstance(decay, str):
        if decay == "linear":
            return linear_size
        if decay == "exponential":
            return exponential_size
        if decay.startswith("poly"):
            power = int(decay[-1])
            return lambda size: size**power
        raise ValueError(f"Unknown decay type {decay}")
    if callable(decay):
        return decay

    raise ValueError("size_decay has to be either a string or callable")


def get_dia_decay(decay: Union[str, Callable[[int], float]]) -> Callable[[int], float]:
    if isinstance(decay, str):
        if decay == "linear":
            return linear_dia
        if decay == "exponential":
            return exponential_dia
        if decay.startswith("poly"):
            power = int(decay[-1])
            return lambda dia: dia**power
        raise ValueError(f"Unknown decay type {decay}")
    if callable(decay):
        return decay

    raise ValueError("dia_decay has to be either a string or callable")


def random_cv_hyper_opt(
    phys_ridge: PhysicalRidge,
    params: Dict,
    X: np.ndarray,
    y: np.ndarray,
    cv: int = 5,
    num_trials: int = 100,
    groups: Sequence[int] = (),
) -> Dict:
    """
    Estimate the hyper parameters of the Physical Ridge by random search.

    :param phys_ridge: Instance of the physical ridge class

    :param params: Dictionary with candiates for all parameters. Example:
        {
            'lamb_dia': [1e-12, 1e-11, 1e-4, 1e-3],
            'lamb_size': [1e-12, 1e-11, 1e-4, 1e-3],
            'size_decay': ['linear', 'exponential'],
            'dia_decay': ['linear', 'exponential']
        }
        on each iteration a random combination of the the specified
        candidates will be attempted.

    :param cv: Number of folds used for cross validation

    :param num_trials: Number of combinations of hyper parameters
        that will be tried

    :param groups: Grouping information for X and y matrix. See docstring
        of `clease.tools.split_dataset` for furhter information.
    """
    # pylint: disable=too-many-locals
    best_param = None
    best_cv = 0.0
    best_mse = 0.0
    best_coeff = None

    cv_params = []
    last_print = time.time()
    partitions = split_dataset(X, y, nsplits=cv, groups=groups)

    for i in range(num_trials):
        lamb_dia = choice(params.get("lamb_dia", [phys_ridge.lamb_dia]))
        lamb_size = choice(params.get("lamb_size", [phys_ridge.lamb_size]))
        size_decay = choice(params.get("size_decay", [phys_ridge.size_decay]))
        dia_decay = choice(params.get("dia_decay", [phys_ridge.dia_decay]))

        phys_ridge.lamb_dia = lamb_dia
        phys_ridge.lamb_size = lamb_size
        phys_ridge.size_decay = size_decay
        phys_ridge.dia_decay = dia_decay

        param_dict = {
            "lamb_dia": lamb_dia,
            "lamb_size": lamb_size,
            "size_decay": size_decay,
            "dia_decay": dia_decay,
        }

        cv_score = 0.0
        mse = 0.0
        for p in partitions:
            coeff = phys_ridge.fit(p["train_X"], p["train_y"])
            pred = p["validate_X"].dot(coeff)
            cv_score += np.sqrt(np.mean((pred - p["validate_y"]) ** 2))

            pred = p["train_X"].dot(coeff)
            mse += np.sqrt(np.mean((pred - p["train_y"]) ** 2))

        cv_score /= len(partitions)
        mse /= len(partitions)

        cv_params.append((cv_score, param_dict))
        if best_param is None or cv_score < best_cv:
            best_cv = cv_score
            best_mse = mse
            best_coeff = coeff
            best_param = param_dict

        if time.time() - last_print > 30:
            msg = (
                f"{i} of {num_trials}. CV: {best_cv*1000.0} meV/atom. "
                f"MSE: {best_mse*1000.0} meV/atom. Params: {best_param}"
            )
            logger.info(msg)
            last_print = time.time()

    cv_params = sorted(cv_params, key=lambda x: x[0])
    res = {
        "best_coeffs": best_coeff,
        "best_params": best_param,
        "best_cv": best_cv,
        "cvs": [x[0] for x in cv_params],
        "params": [x[1] for x in cv_params],
    }
    return res



from typing import List, Sequence, Optional
from ase import Atoms

__all__ = ("AtomsManager",)


class AtomsManager:
    """
    Manager class for the Atoms object used in a cluster expansion context.
    This class can return indices of the Atoms object grouped according to
    various schemes.

    :param atoms: ASE Atoms object
    """

    def __init__(self, atoms: Atoms = None) -> None:
        self.atoms = atoms

    @property
    def atoms(self) -> Atoms:
        return self._atoms

    @atoms.setter
    def atoms(self, other: Optional[Atoms]) -> None:
        if other is not None and not isinstance(other, Atoms):
            raise TypeError(
                (
                    "Trying to set atoms with a non-atoms object. "
                    f"Expected atoms, got {type(other)}"
                )
            )
        self._atoms = other

    def __eq__(self, other: "AtomsManager") -> bool:
        if not isinstance(other, AtomsManager):
            return NotImplemented

        return self.atoms == other.atoms

    def index_by_tag(self) -> List[List[int]]:
        """Return atomic indices that are grouped by their tags.

        This method assumes that all atoms are tagged and the tags are in a
        continuous sequence starting from 0.
        """

        tags = sorted(set(atom.tag for atom in self.atoms))
        ind_by_tag = [[] for _ in tags]

        for atom in self.atoms:
            ind_by_tag[atom.tag].append(atom.index)
        return ind_by_tag

    def index_by_symbol(self, symbols: List) -> List[List[int]]:
        """Group atomic indices by its atomic symbols.

        Example:

            If symbols = ['Au', 'Cu'], then indices where the indices of Au are
            returned as group 1 (first list in the nested list of indices) and
            the indices of Cu are returned as group 2 (second list in the
            nested list of indices).

            If symbols = ['Au, ['Cu', 'X'], 'Ag'], then indices of Au are
            returned as group 1, the indices of Cu and X are returned as group
            2 and the indices of Ag are returned as group 3.

        :param symbols: List of symbols that define a group
        """
        ind_by_symbol = [[] for _ in symbols]
        group_map = {}
        for i, item in enumerate(symbols):
            if isinstance(item, list):
                for x in item:
                    group_map[x] = i
            else:
                group_map[item] = i

        # Loop over indices
        for atom in self.atoms:
            ind_by_symbol[group_map[atom.symbol]].append(atom.index)
        return ind_by_symbol

    def unique_elements(self, ignore: Sequence[str] = ()) -> List[str]:
        """Return a list of symbols of unique elements.

        :param ignore: List of symbols to ignore in finding unique elements.
        """
        all_unique = set(self.atoms.symbols)
        return list(all_unique - set(ignore))

    def single_element_sites(self, allowed_elements: List[List[str]]) -> List[int]:
        """
        Return a list of sites that can only be occupied by a single element
        according to allowed_elements.

        :param allowed_elements: List of allowed elements on each site.
            `allowed_elements` takes the same format/style as
            `basis_elements` in settings (i.e., a nested list with each
            sublist containing a list of elements in a basis). It is assumed
            that the first elements in each group is present in self.atoms.
            For example, if `allowed_elements` is
            [['Au', 'Ag' 'X], ['Cu', 'X']] it means that all sites of the
            `self.atoms` must be occupied by either Au or Cu. All of the
            original sites occupied by Au can be occupied by Ag or X in cluster
            expansion, and all of the original sites occupied by Cu can be
            occupied by Cu or X in the cluster expansion.
        """
        single_site_symb = [x[0] for x in allowed_elements if len(x) == 1]
        single_sites = []
        for atom in self.atoms:
            if atom.symbol in single_site_symb:
                single_sites.append(atom.index)
        return single_sites



"""Calculator for Cluster Expansion."""
import sys
import contextlib
from typing import Dict, Optional, TextIO, Union, List
import numpy as np
from ase import Atoms
from ase.calculators.calculator import Calculator
from clease_cxx import PyCEUpdater
from clease.datastructures import SystemChange, SystemChanges
from clease.corr_func import CorrFunction
from clease.settings import ClusterExpansionSettings


class MovedIgnoredAtomError(Exception):
    """Raised when ignored atoms is moved."""


class KeepChanges:
    def __init__(self):
        self.keep_changes = True


# pylint: disable=too-many-instance-attributes
class Clease(Calculator):
    """Class for calculating energy using CLEASE.

    :param settings: `ClusterExpansionSettings` object

    :param eci: Dictionary containing cluster names and their ECI values

    :param init_cf: (Optional) One can supply the correlation function values
        of the Atoms object to skip the initial assessment step. The dictionary
        should contain cluster names (same as the ones provided in `eci`) and
        their correlation function values.

    :logfile: One can pass the file object or string of the file name to get a
        log file. Do not specify or set it to *None* to avoid generating a log
        file. Use '-' for stdout.
    """

    name = "CLEASE"
    implemented_properties = ["energy"]

    def __init__(
        self,
        settings: ClusterExpansionSettings,
        eci: Dict[str, float],
        init_cf: Optional[Dict[str, float]] = None,
        logfile: Union[TextIO, str, None] = None,
    ) -> None:
        Calculator.__init__(self)

        if not isinstance(settings, ClusterExpansionSettings):
            msg = "settings must be CEBulk or CECrystal object."
            raise TypeError(msg)
        self.parameters["eci"] = eci
        self.settings = settings
        self.corrFunc = CorrFunction(settings)
        self.eci = eci
        # store cluster names
        self.cf_names = list(eci.keys())

        # calculate init_cf or convert init_cf to array
        if init_cf is None or isinstance(init_cf, dict):
            self.init_cf = init_cf
        else:
            msg = "'init_cf' must be a dictionary containing cluster names "
            msg += "and their correlation function values."
            raise TypeError(msg)

        if self.init_cf is not None and len(self.eci) != len(self.init_cf):
            msg = "length of provided ECIs and correlation functions do not "
            msg += "match"
            raise ValueError(msg)

        # logfile
        if isinstance(logfile, str):
            if logfile == "-":
                logfile = sys.stdout
            else:
                # pylint: disable=consider-using-with
                logfile = open(logfile, "a")
        self.logfile = logfile

        # reference atoms for calculating the cf and energy for new atoms
        self.atoms = None
        self.symmetry_group = None
        self.is_backround_index = None

        # C++ updater initialized when atoms are set
        self.updater = None

    def _set_norm_factors(self) -> None:
        """Set normalization factor for each cluster.

        The normalization factor only kicks in when the cell is too small and
        thus, include self-interactions. This methods corrects the impact of
        self-interactions.
        """
        symm_group = np.zeros(len(self.settings.atoms), dtype=np.uint8)
        for num, group in enumerate(self.settings.index_by_sublattice):
            symm_group[group] = num

        cluster_list = self.settings.cluster_list
        for cluster in cluster_list:
            fig_keys = list(set(cluster.get_all_figure_keys()))
            num_occ = {}
            for key in fig_keys:
                num_occ[key] = cluster_list.num_occ_figure(
                    key, cluster.name, symm_group, self.settings.trans_matrix
                )
            num_fig_occ = cluster.num_fig_occurences
            norm_factors = {}
            for key in fig_keys:
                tot_num = num_occ[key]
                num_unique = len(set(key.split("-")))
                norm_factors[key] = float(tot_num) / (num_unique * num_fig_occ[key])

            norm_factor_list = []
            for fig in cluster.indices:
                key = cluster.get_figure_key(fig)
                norm_factor_list.append(norm_factors[key])
            cluster.info["normalization_factor"] = norm_factor_list

    def set_atoms(self, atoms: Atoms) -> None:
        self.atoms = atoms

        if self.init_cf is None:
            self.init_cf = self.corrFunc.get_cf_by_names(self.atoms, self.cf_names)

        if len(self.settings.atoms) != len(atoms):
            msg = "Passed Atoms object and settings.atoms should have same "
            msg += "number of atoms."
            raise ValueError(msg)

        if not np.allclose(atoms.positions, self.settings.atoms.positions):
            msg = "Positions of all atoms in the passed Atoms object and "
            msg += "settings.atoms should be the same. "
            raise ValueError(msg)

        self.symmetry_group = np.zeros(len(atoms), dtype=int)
        for symm, indices in enumerate(self.settings.index_by_sublattice):
            self.symmetry_group[indices] = symm
        self.is_backround_index = np.zeros(len(atoms), dtype=np.uint8)
        self.is_backround_index[self.settings.background_indices] = 1

        self._set_norm_factors()
        self.updater = PyCEUpdater(
            self.atoms,
            self.settings,
            self.init_cf,
            self.eci,
            self.settings.cluster_list,
        )

    def get_energy_given_change(self, system_changes: SystemChanges, keep_changes=False) -> float:
        """
        Calculate the energy when the change is known. No checking will be
        performed.

        :param system_changes: List of system changes. For example, if the
            occupation of the atomic index 23 is changed from Mg to Al,
            system_change = [(23, Mg, Al)]. If an Mg atom occupying the atomic
            index 26 is swapped with an Al atom occupying the atomic index 12,
            system_change = [(26, Mg, Al), (12, Al, Mg)]
        :param keep_changes: Should the calculator revert to the state prior to
            the system changes, or should the calculator stay in the new state
            with the given system changes. Default: False
        """
        with self.with_system_changes(system_changes) as keeper:
            keeper.keep_changes = keep_changes
            return self.energy

    def check_state(self, atoms: Atoms, tol: float = 1e-15):
        res = super().check_state(atoms)
        syst_ch = self.indices_of_changed_atoms

        if syst_ch and "energy" not in res:
            res.append("energy")
        return res

    def reset(self):
        self.results = {}

    # pylint: disable=signature-differs
    def calculate(
        self,
        atoms: Atoms,
        properties: Union[List[str], str],
        system_changes: SystemChanges,
    ) -> float:
        """Calculate the energy of the passed Atoms object.

        If accept=True, the most recently used atoms object is used as a
        reference structure to calculate the energy of the passed atoms.
        Returns energy.

        :param atoms: ASE Atoms object

        :param properties: List of what needs to be calculated.
            It can only by 'energy' at the moment.

        :param system_changes: List of system changes. For example, if the
            occupation of the atomic index 23 is changed from Mg to Al,
            system_change = [(23, Mg, Al)]. If an Mg atom occupying the atomic
            index 26 is swapped with an Al atom occupying the atomic index 12,
            system_change = [(26, Mg, Al), (12, Al, Mg)]
        """
        self.update_energy()
        self.energy = self.updater.get_energy()
        return self.energy

    def clear_history(self) -> None:
        self.updater.clear_history()

    def restore(self) -> None:
        """Restore the Atoms object to its original configuration and energy.

        This method reverts the Atoms object to its oldest state stored in
        memory. The state is restored to either
        (1) an initial state when the calculator was attached, or
        (2) the state at which the `clear_history()` method was invoked last
            time.

        NOTE: The maximum capacity for the history buffer is 1000 steps
        """
        self.updater.undo_changes()
        self.energy = self.updater.get_energy()

    def update_energy(self) -> None:
        """Update correlation function and get new energy."""
        self.update_cf()
        self.energy = self.updater.get_energy()

    @property
    def energy(self):
        return self.results.get("energy", None)

    @energy.setter
    def energy(self, value):
        self.results["energy"] = value

    @property
    def indices_of_changed_atoms(self) -> List[int]:
        """Return the indices of atoms that have been changed."""
        changed = self.updater.get_changed_sites(self.atoms)
        for index in changed:
            if self.is_backround_index[index] and not self.settings.include_background_atoms:
                msg = f"Atom with index {index} is a background atom."
                raise MovedIgnoredAtomError(msg)

        return changed

    def get_cf(self) -> Dict[str, float]:
        """Return the correlation functions as a dict"""
        return self.updater.get_cf()

    def update_cf(self, system_changes: SystemChanges = None) -> None:
        """Update correlation function based on the reference value.

        :param system_changes: List of system changes. For example, if the
            occupation of the atomic index 23 is changed from Mg to Al,
            system_change = [(23, Mg, Al)]. If an Mg atom occupying the atomic
            index 26 is swapped with an Al atom occupying the atomic index 12,
            system_change = [(26, Mg, Al), (12, Al, Mg)]
            If system_changes is None, the correlation functions are recalculated.
            Default: None.
        """
        if system_changes is None:
            swapped_indices = self.indices_of_changed_atoms
            symbols = self.updater.get_symbols()
            system_changes = [
                SystemChange(
                    index=idx,
                    old_symb=symbols[idx],
                    new_symb=self.atoms[idx].symbol,
                    name="internal_symbol_change",
                )
                for idx in swapped_indices
            ]
        for change in system_changes:
            self.updater.update_cf(change)

    @property
    def cf(self) -> List[float]:
        temp_cf = self.updater.get_cf()
        return [temp_cf[x] for x in self.cf_names]

    def log(self) -> None:
        """Write energy to log file."""
        if self.logfile is None:
            return
        self.logfile.write(f"{self.energy}\n")
        self.logfile.flush()

    def update_eci(self, eci: Dict[str, float]) -> None:
        """Update the ECI values.

        :param eci: dictionary with new ECI values
        """
        self.eci = eci
        self.updater.set_eci(eci)
        self._on_eci_changed()

    def get_singlets(self) -> np.ndarray:
        return self.updater.get_singlets()

    def get_energy(self) -> float:
        self.energy = self.updater.get_energy()
        return self.energy

    def _on_eci_changed(self):
        """
        Callback that is called after ECIs are changed. It is provided such
        that calculators inheriting from this class (and implementing this
        method) can be notified when the ECIs are changed.
        """

    @contextlib.contextmanager
    def with_system_changes(self, system_changes: SystemChanges):
        """
        : param system_changes: List of system changes. For example, if the
            occupation of the atomic index 23 is changed from Mg to Al,
            system_change = [(23, Mg, Al)]. If an Mg atom occupying the atomic
            index 26 is swapped with an Al atom occupying the atomic index 12,
            system_change = [(26, Mg, Al), (12, Al, Mg)]
        """
        keeper = KeepChanges()  # We need an object we can mutate to flag for cleanup
        try:
            # Apply the updates
            self.apply_system_changes(system_changes)
            yield keeper
        finally:
            if keeper.keep_changes:
                # Keep changes
                self.keep_system_changes()
            else:
                # Revert changes
                self.undo_system_changes()

    def apply_system_changes(self, system_changes: SystemChanges) -> None:
        """Apply a set of changes to the calculator, and evaluate the energy"""
        self.reset()
        # Apply the updates
        self.update_cf(system_changes)
        self.energy = self.updater.get_energy()

    def undo_system_changes(self) -> None:
        """Revert a set of changes. The changes passed in should be the original
        sequence of system changes which were applied. Restores the original results."""
        self.restore()  # Also restores results

    def keep_system_changes(self) -> None:
        """A set of system changes are to be kept. Perform necessary actions to prepare
        for a new evaluation."""
        self.clear_history()



from typing import Union
from abc import ABC
from ase import Atoms
from .mc_evaluator import MCEvaluator, construct_evaluator

__all__ = ("BaseMC",)


class BaseMC(ABC):
    """Base Monte Carlo Class.
    Initializes the internal atoms and evaluator objects.

    Args:
        system (Union[Atoms, MCEvaluator]): Either an ASE Atoms object
            with an attached calculator, or a pre-initialized
            :class:`~clease.montecarlo.mc_evaluator.MCEvaluator`
            object.
    """

    def __init__(
        self,
        system: Union[Atoms, MCEvaluator],
    ):

        self.evaluator = system

    @property
    def evaluator(self) -> MCEvaluator:
        """The internal evaluator object.

        :getter: Returns the internal
            :class:`~clease.montecarlo.mc_evaluator.MCEvaluator` object.
        :setter: Sets the internal evaluator object. Can either accept an
            atoms object, or a pre-initialized evaluator object.
            See ``system`` in the docstring of the class constructor.
        :type: :class:`~clease.montecarlo.mc_evaluator.MCEvaluator`
        """
        return self._evaluator

    @evaluator.setter
    def evaluator(self, value: Union[Atoms, MCEvaluator]) -> None:
        """Set the evaluator object. If the value is an Atoms object,
        the evaluator is created on basis of the attached calculator object.
        """
        self._evaluator = construct_evaluator(value)

    @property
    def atoms(self) -> Atoms:
        """The internal Atoms object.

        :type: :class:`ase.atoms.Atoms`
        """
        return self.evaluator.atoms



# pylint: disable=undefined-variable
import logging
from deprecated import deprecated
from .version import *

# Do imports in the proper order, avoid cyclic imports
from . import datastructures
from . import basis_function
from . import logging_utils
from . import mp_logger
from . import cluster
from . import jsonio
from . import tools
from . import db_util
from . import corr_func
from . import settings
from . import calculator
from . import montecarlo
from . import structgen
from . import concentration

from .convexhull import *
from .data_manager import *
from .structure_mapper import *
from .structgen.new_struct import NewStructures  # Promotion for convenience
from .svd import *
from .concentration import *
from .regression_old import *
from .cluster_coverage import *
from .logging_utils import *
from .evaluate import *

# Import this global, so it can be disabled with "clease.REQUIRE_COMPATIBLE_TABLE_VERISON = False"
from .db_util import toggle_require_compatible_table_version

logger = logging.getLogger(__name__)


@deprecated(version="0.10.0", reason="import CEBulk from clease.settings instead")
def CEBulk(*args, **kwargs):
    return settings.CEBulk(*args, **kwargs)


@deprecated(version="0.10.0", reason="import CECrystal from clease.settings instead")
def CECrystal(*args, **kwargs):
    return settings.CECrystal(*args, **kwargs)


@deprecated(version="0.10.0", reason="import CESlab from clease.settings instead")
def CESlab(*args, **kwargs):
    return settings.CESlab(*args, **kwargs)


ADDITIONAL = (
    "__version__",
    "version",
    "settings",
    "basis_function",
    "jsonio",
    "corr_func",
    "structgen",
    "NewStructures",
    "cluster",
    "tools",
    "montecarlo",
    "mp_logger",
    "CEBulk",
    "CECrystal",
    "CESlab",
    "concentration",
    "db_util",
    "calculator",
    "toggle_require_compatible_table_version",
    "logging_utils",
    "datastructures",
)

__all__ = (
    version.__all__
    + concentration.__all__
    + evaluate.__all__
    + convexhull.__all__
    + data_manager.__all__
    + structure_mapper.__all__
    + svd.__all__
    + regression_old.__all__
    + cluster_coverage.__all__
    + logging_utils.__all__
) + ADDITIONAL



"""This module defines the "Figure" class, which is a collection of FourVector objects."""
from typing import Iterable, Any, Tuple
import attr
import ase
import numpy as np
from .four_vector import FourVector

__all__ = ("Figure",)


def _convert_figure(x: Any) -> Any:
    """Perform a possible type conversion on the input, prior to validation.
    We allow any iterable, which can be converted into a tuple.
    No duplication checks are made."""
    # Fast-track the tuple case, since this is the correct type of the class
    # and it is also an Iterable.
    if not isinstance(x, tuple) and isinstance(x, Iterable):
        # Try to convert into a tuple
        x = tuple(x)
    return x


@attr.s(frozen=True, order=False)
class Figure:
    """Class which defines a Figure, i.e. a collection of FourVector objects
    which defines a single cluster. Each entry in the components must be a FourVector,
    and is checked upon construction. The order of the FourVector objects is preserved,
    and is not checked for duplicate entries.

    It is possible to pass in the FourVectors in any Iterable, which will be then
    converted into a tuple, e.g.

    >>> from clease.datastructures import Figure, FourVector
    >>> fv1 = FourVector(0, 0, 0, 0)
    >>> fv2 = FourVector(1, 1, 1, 1)
    >>> Figure([fv1, fv2])
    >>> Figure((fv1, fv2))
    """

    components: Tuple[FourVector] = attr.ib(
        converter=_convert_figure, validator=attr.validators.instance_of(tuple)
    )

    @components.validator
    def _validate_all_four_vectors(self, attribute, value):
        """Perform a check that all elements in the components sequence are FourVector objects"""
        # pylint: disable=unused-argument, no-self-use
        # The signature of this function is dictated by attrs.
        for ii, v in enumerate(value):
            if not isinstance(v, FourVector):
                raise TypeError(
                    f"All values must FourVector type, got {value} "
                    f"of type {type(v)} in index {ii}."
                )

    def to_cartesian(self, prim: ase.Atoms, transposed_cell: np.ndarray = None) -> np.ndarray:
        """Get the Figure in terms of the cartesian coordinates, as defined
        by the primitive lattice.

        Args:
            prim (ase.Atoms): The primitive atoms cell, which defines the lattice.
            transposed_cell (np.ndarray, optional): Optional pre-calculated transposed
             cell of the primitive.

        Returns:
            np.ndarray: The cartesian coordinates of the Figure.
        """
        if transposed_cell is None:
            transposed_cell = prim.get_cell().T

        cart = np.zeros((self.size, 3))
        for ii, fv in enumerate(self.components):
            cart[ii, :] = fv.to_cartesian(prim, transposed_cell=transposed_cell)
        return cart

    def get_diameter(self, prim: ase.Atoms, transposed_cell: np.ndarray = None) -> float:
        """Calculate the diameter of the figure, as the maximum distance to the
        geometric center of the figure in cartesian coordinates.

        Args:
            prim (ase.Atoms): The primitive atoms cell, which defines the lattice.
            transposed_cell (np.ndarray, optional): Optional pre-calculated transposed
             cell of the primitive.

        Returns:
            float: The diameter of Figure.
        """

        X = self.to_cartesian(prim, transposed_cell=transposed_cell)
        # Center at the geometric center
        X -= np.mean(X, axis=0)
        # Convert to distances
        dists = np.linalg.norm(X, axis=1, ord=2)
        # Find the maximum distance (radius), and convert to diameter
        return float(2 * dists.max())

    @property
    def size(self) -> int:
        """Size of this cluster, i.e. the number of FourVector components."""
        return len(self.components)



import numpy as np
from scipy.optimize import minimize


class ExponentialFilter:
    def __init__(self, min_time=1, max_time=10, n_subfilters=2, dt=1):
        self.min_time = min_time
        self.max_time = max_time
        self.filter_values = np.zeros(n_subfilters)
        self.tau = np.linspace(min_time, max_time, n_subfilters)
        self.dt = dt
        self.weights = np.exp(-self.dt / self.tau)
        self.n_samples = 0

    def get_normalization_factors(self):
        """Return the normalization factors."""
        norm_factor = 1.0 - np.exp(-self.n_samples * self.dt / self.tau)
        norm_factor /= 1.0 - np.exp(-self.dt / self.tau)
        return norm_factor

    def reset(self):
        """Clear the content."""
        self.filter_values = np.zeros_like(self.filter_values)
        self.n_samples = 0

    def add(self, value):
        """Add a new value."""
        self.filter_values = self.filter_values * self.weights + value
        self.n_samples += 1

    def get(self):
        """Get the filtered values."""
        normalization = self.get_normalization_factors()
        return self.filter_values / normalization

    def get_std(self, std_value):
        """
        Compute the standard deviation of the filtered value
        User has to supply the standard deviation of the individual values
        """
        normalization = self.get_normalization_factors()
        return std_value / np.sqrt(normalization)

    def status_msg(self, std_value=None):
        """Return status message as a string."""
        msg = "======== EXPONENTIAL FILTER INFORMATION ========\n"
        msg += f"Number of samples: {self.n_samples}\n"
        values = self.get()
        std = np.zeros_like(values)
        if std_value is not None:
            std = self.get_std(std_value)
        for i in range(len(self.filter_values)):
            if std_value is None:
                msg += f"Decay time: {self.tau[i]}. "
                msg += f"Value: {values[i]}\n"
            else:
                msg += f"Decay time: {int(self.tau[i])}. "
                msg += f"Value: {values[i]} +- {std[i]} "
                msg += f"({100 * std[i] / values[i]} %)\n"
        msg += "===============================================\n"
        return msg

    def slope(self, std_value=None):
        """Estimate the filter slope."""
        x = self.tau / -2.0
        y = self.get()
        var = 1 / self.tau
        Sx = np.sum(x / var)
        Sy = np.sum(y / var)
        Sxx = np.sum(x**2 / var)
        Sxy = np.sum(x * y / var)
        Ss = np.sum(1.0 / var)
        N = len(y)
        slope = (Ss * Sxy - Sx * Sy) / (Ss * Sxx - Sx**2)
        if std_value is None:
            return slope
        std = self.get_std(std_value)
        var_slope = np.sum(std) / (N * Sxx - Sxy)
        return slope, np.sqrt(var_slope)

    def exponential_extrapolation(self):
        """
        Return the exponential extrapolated value at infinity
        This can be useful in ground state searches as one knows
        at low temperature the energy always decreases and the
        this value can be a measure of the true ground state energy
        """
        x = self.tau / -2.0
        y = self.get()
        x += x[-1]

        def cost(params):
            constant = params[0]
            prefactor = params[1]
            damping = params[2]
            predict = constant + prefactor * np.exp(-damping * x)
            return np.sum((y - predict) ** 2)

        res = minimize(cost, x0=(y[0], 0.0, 0.0))
        params = res["x"]
        return params[0]



"""Module for setting up pseudospins and basis functions."""
from abc import ABC, abstractmethod
import math
from typing import List, Dict, Optional, Sequence
import numpy as np
from clease.jsonio import jsonable
from clease.gramSchmidthMonomials import GramSchmidtMonimial

__all__ = (
    "BasisFunction",
    "Polynomial",
    "Trigonometric",
    "BinaryLinear",
    "basis_function_from_dict",
)


@jsonable("basisfunction")
class BasisFunction(ABC):
    """Base class for all Basis Functions."""

    name = None

    def __init__(self, unique_elements: Sequence[str]) -> None:
        self.unique_elements = unique_elements
        if self.num_unique_elements < 2:
            raise ValueError("Systems must have more than 1 type of element.")
        assert self.name, "Need to set a name"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BasisFunction):
            return False
        return self.name == other.name and self.unique_elements == other.unique_elements

    @property
    def unique_elements(self) -> List[str]:
        return self._unique_elements

    @unique_elements.setter
    def unique_elements(self, elements):
        self._unique_elements = sorted(set(elements))

    @property
    def num_unique_elements(self) -> int:
        return len(self.unique_elements)

    @property
    def spin_dict(self) -> Dict[str, int]:
        return self.get_spin_dict()

    @property
    def basis_functions(self) -> List[Dict[str, float]]:
        return self.get_basis_functions()

    @abstractmethod
    def get_spin_dict(self):
        """Get spin dictionary."""

    @abstractmethod
    def get_basis_functions(self):
        """Get basis function."""

    # pylint: disable=no-self-use
    def customize_full_cluster_name(self, full_cluster_name: str) -> str:
        """Customize the full cluster names. Default is to do nothing."""
        return full_cluster_name

    def todict(self) -> dict:
        """
        Create a dictionary representation of the basis function class
        """
        return {"name": self.name, "unique_elements": self.unique_elements}


class Polynomial(BasisFunction):
    """Pseudospin and basis function from Sanchez et al.

    Sanchez, J. M., Ducastelle, F. and Gratias, D. (1984).
    Generalized cluster description of multicomponent systems.
    Physica A: Statistical Mechanics and Its Applications, 128(1-2), 334-350.
    """

    name = "polynomial"

    def get_spin_dict(self) -> Dict[str, int]:
        """Define pseudospins for all consistuting elements."""
        gram_schmidt = GramSchmidtMonimial(self.num_unique_elements)
        spin_values = gram_schmidt.values
        spin_dict = {}
        for x in range(self.num_unique_elements):
            spin_dict[self.unique_elements[x]] = spin_values[x]
        return spin_dict

    def get_basis_functions(self) -> List[Dict[str, float]]:
        """Create basis functions to guarantee the orthonormality."""
        gram_schmidt = GramSchmidtMonimial(self.num_unique_elements)
        gram_schmidt.build()
        return gram_schmidt.basis_functions(self.unique_elements)


class Trigonometric(BasisFunction):
    """Pseudospin and basis function from van de Walle.

    van de Walle, A. (2009).
    Multicomponent multisublattice alloys, nonconfigurational entropy and other
    additions to the Alloy Theoretic Automated Toolkit. Calphad, 33(2),
    266-278.
    """

    name = "trigonometric"

    def get_spin_dict(self) -> Dict[str, int]:
        """Define pseudospins for all consistuting elements."""
        spin_values = list(range(self.num_unique_elements))
        spin_dict = {}
        for x in range(self.num_unique_elements):
            spin_dict[self.unique_elements[x]] = spin_values[x]
        return spin_dict

    def get_basis_functions(self) -> List[Dict[str, float]]:
        """Create basis functions to guarantee the orthonormality."""
        alpha = list(range(1, self.num_unique_elements))
        bf_list = []

        for a in alpha:
            bf = {}
            for key, value in self.spin_dict.items():
                var = 2 * np.pi * math.ceil(a / 2.0) * value
                var /= self.num_unique_elements
                if a % 2 == 1:
                    bf[key] = -np.cos(var) + 0.0
                else:
                    bf[key] = -np.sin(var) + 0.0

            # normalize the basis function
            sum_ = sum(bf[key] * bf[key] for key in self.spin_dict)
            normalization_factor = np.sqrt(self.num_unique_elements / sum_)

            for key, value in bf.items():
                bf[key] = value * normalization_factor

            bf_list.append(bf)

        return bf_list


def _kronecker(i: int, j: int) -> int:
    """Kronecker delta function."""
    if i == j:
        return 1
    return 0


class BinaryLinear(BasisFunction):
    """Pseudospin and basis function from Zhang and Sluiter.

    Zhang, X. and Sluiter M.
    Cluster expansions for thermodynamics and kinetics of multicomponent
    alloys.
    Journal of Phase Equilibria and Diffusion 37(1) 44-52.
    """

    name = "binary_linear"

    def __init__(self, unique_elements: List[str], redundant_element: Optional[str] = "auto"):
        super().__init__(unique_elements)
        if redundant_element == "auto":
            self.redundant_element = sorted(unique_elements)[0]
        else:
            self.redundant_element = redundant_element

    def get_spin_dict(self) -> Dict[str, int]:
        """Define pseudospins for all consistuting elements."""
        spin_values = list(range(self.num_unique_elements))
        spin_dict = {}
        for x in range(self.num_unique_elements):
            spin_dict[self.unique_elements[x]] = spin_values[x]
        return spin_dict

    def get_basis_functions(self) -> List[Dict[str, float]]:
        """Create orthonormal basis functions.

        Due to the constraint that any site is occupied by exactly one element,
        we only need to track N-1 species if there are N species.
        Hence, the first element specified is redundant, and will not
        have a basis function.
        """
        bf_list = []
        num_bf = self.num_unique_elements
        for bf_num in range(num_bf):
            if self.unique_elements[bf_num] == self.redundant_element:
                continue
            new_bf = {
                symb: float(_kronecker(i, bf_num)) for i, symb in enumerate(self.unique_elements)
            }
            bf_list.append(new_bf)
        return bf_list

    def _decoration2element(self, dec_num: int) -> str:
        """Get the element with its basis function equal to 1."""
        bf = self.basis_functions[dec_num]
        for k, v in bf.items():
            if v == 1:
                return k
        raise ValueError("Did not find any element where the value is 1.")

    def customize_full_cluster_name(self, full_cluster_name: str) -> str:
        """Translate the decoration number to element names."""
        dec = full_cluster_name.rsplit("_", 1)[1]
        name = full_cluster_name.rsplit("_", 1)[0]
        new_dec = ""
        for decnum in dec:
            element = self._decoration2element(int(decnum))
            new_dec += f"{element}"
        return name + "_" + new_dec

    def todict(self) -> dict:
        """
        Creates a dictionary representation of the class
        """
        dct_rep = super().todict()
        dct_rep["redundant_element"] = self.redundant_element
        return dct_rep


def basis_function_from_dict(dct: dict):
    """Load a dictionary representation of a basis function.

    Example:

        >>> unique_elements = ['Au', 'Cu']
        >>> for bf_func in (Polynomial, Trigonometric, BinaryLinear):
        ...    bf = bf_func(unique_elements)
        ...    dct = bf.todict()
        ...    bf_loaded = basis_function_from_dict(dct)
        ...    # Check that the loaded corresponds to the original
        ...    assert type(bf) is type(bf_loaded)
        ...    assert bf.todict() == bf_loaded.todict()
        >>> # It should also work for the redundant element keyword
        >>> bf = BinaryLinear(unique_elements, redundant_element='Au')
        >>> dct = bf.todict()
        >>> bf_loaded = basis_function_from_dict(dct)
        >>> # Check that the loaded corresponds to the original
        >>> assert bf.todict() == bf_loaded.todict()
    """
    basis_functions = {bf.name: bf for bf in (Polynomial, Trigonometric, BinaryLinear)}

    name = dct.pop("name")

    return basis_functions[name](**dct)



from itertools import product
import copy
from typing import List, Dict
import logging

from ase import Atoms

from clease.tools import flatten, dec_string, list2str
from .cluster_generator import ClusterGenerator
from .cluster import Cluster

logger = logging.getLogger(__name__)
__all__ = ("ClusterList",)


class ClusterList:
    def __init__(self):
        self._clusters = []
        # Format of the names cache: {num_bf: names}
        self._all_cf_name_cache: Dict[int, List[str]] = {}

    @property
    def clusters(self) -> List[Cluster]:
        return self._clusters

    def append(self, cluster: Cluster) -> None:
        self.clusters.append(cluster)
        self.clear_cache()

    def clear(self) -> None:
        """Clear the content."""
        self.clusters.clear()
        self.clear_cache()

    def clear_cache(self) -> None:
        """Clear any cached results."""
        self._all_cf_name_cache.clear()

    @property
    def names(self) -> List[str]:
        """Get all names in the cluster list"""
        return [cluster.name for cluster in self.clusters]

    def get_by_name(self, name) -> List[Cluster]:
        return [cluster for cluster in self.clusters if cluster.name == name]

    def get_by_name_and_group(self, name, group) -> Cluster:
        clusters = self.get_by_name(name)
        for c in clusters:
            if c.group == group:
                return c
        msg = f"There are no cluster named {name} and group {group}"
        raise ValueError(msg)

    def get_by_size(self, size) -> List[Cluster]:
        # Return all clusters with a given size
        return [c for c in self.clusters if c.size == size]

    def max_size(self):
        return max(c.size for c in self.clusters)

    def get_by_group(self, group):
        """Return all clusters in a given symmetry group."""
        return [c for c in self.clusters if c.group == group]

    def get_equivalent_clusters(self, cluster):
        """Return equivalent clusters in other symmetry groups."""
        equiv = []
        for c in self.clusters:
            if c == cluster:
                equiv.append(c)
        return equiv

    @staticmethod
    def get_cf_names(cluster: Cluster, num_bf):
        """Return all possible correlation function names.

        Parameters:

        cluster: Cluster
            Instance of cluster class

        num_bf: int
            Number of basis functions
        """
        name = cluster.name
        eq_sites = cluster.equiv_sites
        bf_list = list(range(num_bf))
        cf_names = []
        for dec in product(bf_list, repeat=cluster.size):
            dec_str = dec_string(dec, eq_sites)
            cf_names.append(name + "_" + dec_str)
        return sorted(list(set(cf_names)))

    def get_all_cf_names(self, num_bf: int) -> List[str]:
        """
        Return a list of all correlation function names

        :param num_bf: Number of basis functions
        """
        if num_bf in self._all_cf_name_cache:
            # Finding all the CF names can be rather expensive, so
            # we use cached results if possible
            logger.debug("Found a cached result for all cf names for %s num bf", num_bf)
            return self._all_cf_name_cache[num_bf]
        # We havn't cached this name list yet, so we need to build it
        logger.debug("Building all CF names for %s num bf", num_bf)
        all_cf = self._build_all_cf_names(num_bf)
        self._all_cf_name_cache[num_bf] = all_cf
        return all_cf

    def _build_all_cf_names(self, num_bf: int) -> List[str]:
        if not isinstance(num_bf, int):
            raise TypeError(f"Number of basis functions must be integer, got {num_bf}")
        all_cf_names = []
        for cluster in self.clusters:
            if cluster.name == "c0":
                all_cf_names.append("c0")
            else:
                all_cf_names += self.get_cf_names(cluster, num_bf)
        return sorted(set(all_cf_names))

    def __len__(self) -> int:
        return len(self.clusters)

    def __getitem__(self, index):
        return self.clusters[index]

    def __delitem__(self, i):
        del self.clusters[i]

    def __array__(self):
        return self.clusters

    def __eq__(self, other):
        if len(self.clusters) != len(other.clusters):
            return False

        for c1, c2 in zip(self.clusters, other.clusters):
            if c1 != c2:
                logger.debug("Clusters not equal: %s and %s", c1, c2)
                return False
        return True

    def sort(self):
        """Sort the internal cluster list"""
        self.clusters.sort()
        # We clear this cache, as the ordering has changed, and hence also potentially
        # the ordering of the cached results.
        self.clear_cache()

    def get_sorted_list(self) -> "ClusterList":
        """Get a new instance of the ClusterList which is sorted."""
        new_list = copy.deepcopy(self)
        new_list.sort()
        return new_list

    def tolist(self):
        return self.clusters

    @property
    def unique_indices(self):
        all_indices = set()
        for cluster in self.clusters:
            all_indices.update(flatten(cluster.indices))
        return list(all_indices)

    @property
    def num_symm_groups(self):
        return max(c.group for c in self.clusters)

    @property
    def unique_indices_per_group(self):
        indices_per_group = []
        for i in range(self.num_symm_groups + 1):
            indices_per_group.append(set())
            for c in self.get_by_group(i):
                indices_per_group[i].update(flatten(c.indices))
        return [list(x) for x in indices_per_group]

    def multiplicity_factors(self, num_sites_per_group: List[int]) -> Dict[str, float]:
        mult_factors = {}
        norm = {}
        for cluster in self.clusters:
            current_factor = mult_factors.get(cluster.name, 0)
            current_norm = norm.get(cluster.name, 0)
            current_factor += len(cluster.indices) * num_sites_per_group[cluster.group]
            current_norm += num_sites_per_group[cluster.group]

            mult_factors[cluster.name] = current_factor
            norm[cluster.name] = current_norm

        for k in mult_factors:
            mult_factors[k] /= norm[k]
        return mult_factors

    def get_subclusters(self, cluster: Cluster):
        """Return all all subclusters of the passed cluster in the list."""
        return [c for c in self.clusters if c.is_subcluster(cluster)]

    def get_figures(self, generator: ClusterGenerator) -> List[Atoms]:
        """Get the figures (in their ASE Atoms object representation)"""
        figures: List[Atoms] = []
        self.sort()
        # We want to skip c0 and c1 anyways
        used_names = {"c0", "c1"}
        for cluster in self.clusters:
            if cluster.name in used_names:
                continue
            figure = cluster.get_figure(generator.prim)

            figure.info = {"name": cluster.name}
            used_names.add(cluster.name)
            figures.append(figure)
        return figures

    def get_occurence_counts(self):
        """Count the number of occurences of all figures."""
        occ_counts = []
        for cluster in self.clusters:
            occ_counts.append(cluster.num_fig_occurences)
        return occ_counts

    def num_occ_figure(self, fig_key, c_name, symm_groups, trans_matrix):
        """Determine the number of occurences of a figure across the structure.

        Parameter:

        fig_key: str
            string containing the indices of consituting atomic indices

        c_name: str
            name of the cluster of the figure

        symm_groups: list
            list containing the symmetry group number of the atomic indices

        trans_matrix: list of dicts
            translation matrix
        """
        figure = list(map(int, fig_key.split("-")))
        tot_num = 0
        clusters = self.get_by_name(c_name)
        occ_count = {}
        for cluster in clusters:
            occ_count[cluster.group] = cluster.num_fig_occurences

        for ref in set(figure):
            group = symm_groups[ref]
            cluster = self.get_by_name_and_group(c_name, group)
            corr_figure = cluster.corresponding_figure(ref, figure, trans_matrix)
            corr_fig_key = list2str(corr_figure)
            tot_num += occ_count[cluster.group][corr_fig_key]
        return tot_num

    def make_names_sequential(self):
        """
        Confirm that cluster names are sequential. If clusters are
        removed from the list, that might not be the case.
        """
        for s in range(2, self.max_size() + 1):
            names = list(set(c.name for c in self.get_by_size(s)))
            names.sort()
            prefixes = list(set(n.rpartition("_")[0] for n in names))
            prefixes.sort()
            name_map = {}

            for n in names:
                # Fix distance string
                new_name = copy.deepcopy(n)
                prefix = n.rpartition("_")[0]
                new_dist = f"{prefixes.index(prefix):04d}"
                new_name = "".join([new_name[:4], new_dist, new_name[8:]])
                name_map[n] = new_name

            prefix_map = {}
            # Fix additional ID
            for k in name_map:
                pref = k.rpartition("_")[0]
                prefix_map[pref] = prefix_map.get(pref, []) + [k]

            for k, v in prefix_map.items():
                v.sort()
                uid_map = {}
                for i, x in enumerate(v):
                    new_name = "".join([x.rpartition("_")[0], "_", str(i)])
                    uid_map[x] = new_name

                for k2, v2 in uid_map.items():
                    name_map[k2] = "".join(
                        [name_map[k2].rpartition("_")[0], "_", v2.rpartition("_")[2]]
                    )
            for c in self.get_by_size(s):
                c.name = name_map[c.name]



# pylint: disable=undefined-variable
from .regression import *
from .generalized_ridge_regression import *
from .bayesian_compressive_sensing import *
from .ga_fit import *
from .constrained_ridge import *
from .physical_ridge import *
from .sequential_cluster_ridge import *

__all__ = (
    regression.__all__
    + generalized_ridge_regression.__all__
    + bayesian_compressive_sensing.__all__
    + ga_fit.__all__
    + physical_ridge.__all__
    + sequential_cluster_ridge.__all__
    + constrained_ridge.__all__
)
